//////////////////////////////////////////////////////////
//
//   - by Bill Carlson
//
//
//////////////////////////////////////////////////////////
#include <afx.h>
#include <windows.h>
#include <ddraw.h>
#include <stdio.h>
#include <time.h>
#include <math.h>


////////////////////////////////////////////////////////////
// Module constants
////////////////////////////////////////////////////////////

#define APPNAME		"WinSkel"
#define CLASSNAME	"WinSkelWndClass"
#define WINDOWNAME	APPNAME

#define		SCREEN_X	1600	// X resolution
#define		SCREEN_Y	1200	// Y resolution

////////////////////////////////////////////////////////////
// Global variables
////////////////////////////////////////////////////////////

char* mapnames[] = {
	"seg000.db",
	"seg001.db",
	"seg005.db",
	"seg006.db",
	"seg007.db",
	"seg009.db" };
int								OnMap;				//Current map being displayed

HWND							g_hWndMain;
HINSTANCE						g_hInstance;
IDirectDraw7					*g_DD7;				//Main DirectDraw Interface
IDirectDrawSurface7				*g_DDPrimarySurface;//Primary Surface (screen)
IDirectDrawSurface7				*g_DDSurface;		//Back Buffer
IDirectDrawSurface7				*g_DDImage;			//Tiles
BOOL							Running = true;		//Global bool, true = the app is not quting, false = the app is quiting
int								view_x;				//Upper left viewport coordinate (x)
int								view_y;				//Upper left viewport coordinate (y)

int								selected_x;			//Currently selected tile (x)
int								selected_y;			//Currently selected tile (y)

//This function is for initializing the DirectDraw stuff
HRESULT InitDDraw(HWND hwnd);

//Called when you quit to clean up DirectDraw and other stuff
VOID	CleanUp( void );

//Loads up the image
VOID	LoadImage( void );

//Draw a frame
HRESULT DrawFrame( void );

//Load up a map
void LoadMap( void );

//Open up the database
void OpenDBase( void );

//Finds the correct terrain piece to place on the tile
int FindTerrain(int piece_id, int type_code);

//Windows procedure call
extern LRESULT CALLBACK MyWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

// This is the structure of each piece of terrain (NOT A MAP TILE!)
struct TERRAIN_STRUCT
{
	unsigned short	tile_id;
	unsigned char	  piece_id;
	unsigned char	  type_code;

	unsigned short	x_origin;	//Left
	unsigned short	y_origin;	//Top
	unsigned short	x_extent;	//Right
	unsigned short	y_extent;	//Bottom

	char				    x_offset;	//Offset when placing x
	char				    y_offset;	//Offset when placing y
	  
	unsigned char	  action;
	unsigned char	  restriction;
	char			      passable;
	char			      underpile;
};
TERRAIN_STRUCT tile[466];

//This is the structure for each map tile
struct TILE_STRUCT
{
	unsigned int	graphic;		//Graphics 1 and 2
	unsigned int	graphic2;		//Graphics 3 and 4
};
TILE_STRUCT		map[255][255];

int FindTerrain(int piece_id, int type_code)
{
	int i;
	if (piece_id == 0 && type_code == 0)
		return 999;
	for(i=0;i<466;i++)
	{
		if ((tile[i].piece_id == piece_id) && (tile[i].type_code == type_code))
			return i;
	}
	return 999;
}

void OpenDBase( void )
{
	FILE* in;
	in = fopen("terrain.db", "rb");
	fread(tile, 1, sizeof(tile), in);
	fclose(in);

	LoadMap();
}

void LoadMap( void )
{
	FILE* in;
	in = fopen(mapnames[OnMap], "rb");
	fread(map, 1, sizeof(map), in);
	fclose(in);
}

////////////////////////////////////////////////////////////
// RegisterMyWndClass
////////////////////////////////////////////////////////////
BOOL RegisterMyWndClass(HINSTANCE	hInst)
{
	WNDCLASS wc;

	wc.style = 0;
	wc.lpfnWndProc = MyWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInst;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = CLASSNAME;

	if (!RegisterClass(&wc))
		return FALSE;

	return TRUE;
}

////////////////////////////////////////////////////////////
// UnregisterMyWndClass
////////////////////////////////////////////////////////////
VOID UnregisterMyWndClass(HINSTANCE	hInst)
{
	UnregisterClass(CLASSNAME, hInst);
}

////////////////////////////////////////////////////////////
// MyWndProc
////////////////////////////////////////////////////////////
LRESULT CALLBACK MyWndProc
(
	HWND	hwnd,
	UINT	msg,
	WPARAM	wParam,
	LPARAM	lParam
)
{
	switch (msg)
	{
		case WM_PAINT:
			{
				HDC			hdc;
				PAINTSTRUCT	ps;

				hdc = BeginPaint(hwnd, &ps);
				EndPaint(hwnd, &ps);

				return 0;
			}
		case WM_ACTIVATEAPP:
			{
				if (wParam)
				{
					// running.. Handle surface loss
					if (g_DDSurface != NULL)
						if (g_DDSurface->IsLost())
							g_DDSurface->Restore();
					if (g_DDPrimarySurface != NULL)
						if (g_DDPrimarySurface->IsLost())
							g_DDPrimarySurface->Restore();
					if (g_DDImage != NULL)
						if (g_DDImage->IsLost())
						{
							g_DDImage->Restore();
							LoadImage();
						}

				}
				else
				{
					// not running
				}
				return 0;
			}
		case WM_CHAR:
			{
				//So this is for taking the input from the keys and doing whatever is necessary...
				switch (wParam) 
				{ 
				case 'a': //Look left                 
					view_x += 55;
					break; 

				case 'd': //Look right
					view_x -= 55;
					break;

				case 'w': //Look up
					view_y += 55;
					break; 

				case 's': //Look down
					view_y -= 55;
					break;

				case 'q': //Go to the previous Map
					OnMap--;
					if (OnMap < 0)
						OnMap = 5;
					LoadMap();
					break;

				case 'e': //Go to the next map
					OnMap++;
					if (OnMap > 5)
						OnMap = 0;
					LoadMap();
					break;
				} 
				return 0;
			}
		case WM_LBUTTONDOWN :
			{
				/*
				I never got around to finishing this, although I probably should.

				The idea was so I could click a tile and tell what its true x,y coord was (since the maps
				are not viewed the same way as they are stored)
				*/
				int x = LOWORD(lParam) - view_x;
				int y = HIWORD(lParam) - view_y;

				selected_x = x / 55;
				selected_y = y / 55;

				if (selected_y > 42)
				{


				}
				else
				{
					//selected_y = 42 - selected_y;
				}
				

				return 0;
			}
		case WM_DESTROY:
			PostQuitMessage(0);
			return 0;

		default:
			break;
	}

	return DefWindowProc(hwnd, msg, wParam, lParam);
}

////////////////////////////////////////////////////////////
// WinMain
////////////////////////////////////////////////////////////

int WINAPI WinMain
(
	HINSTANCE	hInst,
	HINSTANCE	/*hInstPrev*/,
	LPSTR		/*lpstrCmdLine*/,
	int			nCmdShow
)
{
	MSG		msg;
	HWND	hWnd;

	// Store our instance handle in a global variable.
	g_hInstance = hInst;

	// Register our main window class.
	if (!RegisterMyWndClass(hInst))
	{
		MessageBox(NULL, "RegisterMyWndClass failed", APPNAME, MB_OK | MB_ICONEXCLAMATION);
		return 0;
	}

	// Create our main window.
	hWnd = CreateWindow(CLASSNAME,				// Name of window class
						WINDOWNAME,				// Window text (title)
						WS_POPUP,				// Window style 
						CW_USEDEFAULT,			// X coordinate
						CW_USEDEFAULT,			// Y coordinate
						CW_USEDEFAULT,			// Width
						CW_USEDEFAULT,			// Height
						NULL,					// No parent window
						NULL,					// No menu
						hInst,					// Our app's HINSTANCE
						NULL);					// No creation data


	if (hWnd == NULL)
	{
		MessageBox(NULL, "CreateWindow (for main window) failed", APPNAME, MB_OK | MB_ICONEXCLAMATION);
		return 0;
	}

	// Store our main window handle in a global variable.
	g_hWndMain = hWnd;

	// Show our main window.
	ShowWindow(hWnd, nCmdShow);

	//Seed the random number generator
	srand( time( NULL ) );

	//Open up the terrain database and load the first map
	OpenDBase();

	//Load up directdraw
	if (FAILED(InitDDraw(hWnd)))
	{
		// Something failed while initializing
		return FALSE;
	}

	// Main message loop.
	while (Running)
	{
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if (msg.message == WM_QUIT)
				Running = false;
			else
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}

		if (Running)
		{
			DrawFrame();

			//This sleep call isnt necessary, but I figured I wouldnt be a complete
			// resource hog with this demo... In a real game this would be a no-no(usually)
			Sleep(10);
		}
	}

	CleanUp();

	// Clean up: unregister our main window class.
	UnregisterMyWndClass(hInst);

	// Return with the standard return code (from PostQuitMessage).
	return msg.wParam;
}

//This loads the palette so I can use 8-bit color transparency
IDirectDrawPalette * DDLoadPalette(IDirectDraw7 *pdd, LPCSTR szBitmap)
{
    IDirectDrawPalette* ddpal;
    int i;
    int n;
    int fh;
    PALETTEENTRY ape[256];

    if (szBitmap && (fh = _lopen(szBitmap, OF_READ)) != -1)
    {
        BITMAPFILEHEADER bf;
        BITMAPINFOHEADER bi;

        _lread(fh, &bf, sizeof(bf));
        _lread(fh, &bi, sizeof(bi));
        _lread(fh, ape, sizeof(ape));
        _lclose(fh);

        if (bi.biSize != sizeof(BITMAPINFOHEADER))
            n = 0;
        else if (bi.biBitCount > 8)
            n = 0;
        else if (bi.biClrUsed == 0)
            n = 1 << bi.biBitCount;
        else
            n = bi.biClrUsed;

      //
      // a DIB color table has its colors stored BGR not RGB
      // so flip them around.
      //
        for(i=0; i<n; i++ )
        {
            BYTE r = ape[i].peRed;
            ape[i].peRed = ape[i].peBlue;
            ape[i].peBlue = r;
        }
    }
    if (pdd->CreatePalette(DDPCAPS_8BIT, ape, &ddpal, NULL) != DD_OK)
    {
        return NULL;
    } else {
        return ddpal;
    }
}


HRESULT InitDDraw(HWND hwnd)
{
	HRESULT				hr;
	DDSURFACEDESC2		ddsd, ddsd2;
	//DDPIXELFORMAT		ddpf;
	DDSCAPS2			ddsCaps;


	if (FAILED(hr = DirectDrawCreateEx(NULL, (void **)&g_DD7, IID_IDirectDraw7, NULL)))
	{
		//Could not create DDraw7 Object
		return hr;
	}

	if (FAILED(hr = g_DD7->SetCooperativeLevel(hwnd, DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE)))
	{
		//Couldn't set cooperative level for some reason.
		return hr;
	}

	if (FAILED(hr = g_DD7->SetDisplayMode(SCREEN_X, SCREEN_Y, 8, 0, 0)))
	{
		// Couldn't set display mode as requested. CAN'T CONTINUE!
		return hr;
	}


	// Create the primary surface. First initialize our DDSURFACEDESC2
	// struct properly.
	ddsd.dwSize = sizeof(ddsd);   // or, equivalently, sizeof(DDSURFACEDESC2)
	ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;     // 'ddsCaps' member is valid
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
	ddsd.ddsCaps.dwCaps2 = ddsd.ddsCaps.dwCaps3 = ddsd.ddsCaps.dwCaps4 = 0;
	ddsd.dwBackBufferCount = 1;
	
	// Now create the primary surface.
	if (FAILED(hr = g_DD7->CreateSurface(&ddsd, &g_DDPrimarySurface, NULL)))
	{
		// Failed. Handle error somehow.
		return hr;
	}

	IDirectDrawPalette *lpDDPal = DDLoadPalette(g_DD7, "tiles.bmp"); // Call the function above to load the palette

	if (lpDDPal)
		g_DDPrimarySurface->SetPalette(lpDDPal); // this sets the palette for the primary surface


	// Create backbuffer surface
	ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;
	ddsCaps.dwCaps2 = ddsCaps.dwCaps3 = ddsCaps.dwCaps4 = 0;
	if (FAILED(hr = g_DDPrimarySurface->GetAttachedSurface(&ddsCaps, &g_DDSurface)))
	{
		// GetAttachedSurface failed for some reason.
		return hr;
	}

	ddsd2.dwSize = sizeof(ddsd2);
	ddsd2.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;// | DDSD_CKSRCBLT;
	ddsd2.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsd2.ddsCaps.dwCaps2 = ddsd2.ddsCaps.dwCaps3 = ddsd2.ddsCaps.dwCaps4 = 0;
	ddsd2.dwWidth = 1300;
	ddsd2.dwHeight = 3900;
	//ddsd2.ddckCKSrcBlt.dwColorSpaceLowValue = 0xFF00FF;
	//ddsd2.ddckCKSrcBlt.dwColorSpaceHighValue = 0xFF00FF;
	if (FAILED(hr = g_DD7->CreateSurface(&ddsd2, &g_DDImage, NULL)))
	{
		// CreateSurface failed. Handle error or bail out.
		return hr;
	}
	LoadImage();

	//Do the transparency thing...
	DDCOLORKEY ddck;
	ddck.dwColorSpaceLowValue = 142;
	ddck.dwColorSpaceHighValue = 142;
	g_DDImage->SetColorKey( DDCKEY_SRCBLT, &ddck );

	return hr;
}


VOID	CleanUp( void )
{
	g_DD7->Release();
	g_DD7 = NULL;

	g_DDImage->Release();
	g_DDImage = NULL;

	g_DDSurface->Release();
	g_DDSurface = NULL;

	g_DDPrimarySurface->Release();
	g_DDPrimarySurface = NULL;
}

VOID LoadImage( void )
{
	HBITMAP hbitmap, hbmPrev;
	HDC		dc;
	HDC		dc2;

	g_DDImage->GetDC( &dc );
	dc2 = CreateCompatibleDC( dc );

	hbitmap = (HBITMAP)LoadImage(NULL,
                             "tiles.bmp",
                             IMAGE_BITMAP,
                             0,
                             0,
                             LR_CREATEDIBSECTION | LR_LOADFROMFILE);

	hbmPrev = (HBITMAP)SelectObject(dc2, hbitmap);

	BitBlt(dc, 0, 0, 1300, 3900, dc2, 0, 0, SRCCOPY);

	SelectObject(dc2, hbmPrev);
	DeleteObject(hbitmap);
    DeleteDC(dc2);

	g_DDImage->ReleaseDC( dc );

}

HRESULT DrawFrame( void )
{
	HDC					dc;
	DDBLTFX				ddBltFX;
	RECT				rectFill;

	ddBltFX.dwSize = sizeof(ddBltFX);
	ddBltFX.dwFillColor = 0x0;

	rectFill.left = 0;
	rectFill.top = 0;
	rectFill.right = SCREEN_X;
	rectFill.bottom = SCREEN_Y;

	// Make the background black (or whatever the first palette entry is)
	g_DDSurface->Blt(&rectFill, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddBltFX);

	RECT rect[4], rectdst[4];
	int x, y, i;
	int piece_id[4], type_code[4], graphic[4], draworder[4];

	/*
	Note from Bill:

    This code is highly UNOPTIMIZED, in fact its downright slow.  I should be shot for
	writing such a hack... But it works. :)

    The loop could be cut down big time to reduce the multiplications in the first and 
	second if statements...  Also I could probably optimize the storage of the 8-bytes-per-tile
	structure so I wouldnt have to do the &'ing and shifting, but honestly its not too bad...

    Its that or be doing off-boundry calculations which are probably slower.

    Also the FindTerrain function is a bruteforce search method (through 466 items!!!), and its called
	FOUR TIMES on each tile, so thats *bad*.  Its the databases fault, though...  I could really optimize
	how those are stored so it would store them by type_code so the searching would be faster or some
	other ways.  But again, it does work!!

    My next problem is with the draw ordering...  Its pretty jumpled wierd looking code... I was tired 
	and I'm sorry about that.  The basic idea is to find the lowest typecode and draw it, then find the 
	next one and draw it and so on.  This works for 99% of all the tiles, but not for the fountain and a 
	few other instances.  I'm not sure if I'm retarded or if the databases are wrong (they are pretty 
	old anyway).  I'm personally thinking the databases are just wrong... Eitherway it woudl be really
	easy to change the data to make them draw correctly.

  NOTE:
    The draw ordering problem has been fixed, they store them in a very weird way, by having the typecode go to a look-up
    table, so thats how we do it now. (It sucks though)
    */

	for(y=0;y<86;y++)
	{
		for(x=0;x<145;x++)
		{
			if (x*55+view_x >= 0 && (x)*55+view_x <= SCREEN_X)
				if (y*55+view_y >= 0 && (y)*55+view_y <= SCREEN_Y)
					{
						//Grab the two tiles from graphic2
						piece_id[0] = (map[y][x].graphic2 & 0xFF000000) >> 24;
						type_code[0] = (map[y][x].graphic2 & 0x00FF0000) >> 16;
						piece_id[1] = (map[y][x].graphic2 & 0x0000FF00) >> 8;
						type_code[1] = (map[y][x].graphic2 & 0x000000FF);

						//Grab the two tiles from graphic
						piece_id[2] = (map[y][x].graphic & 0xFF000000) >> 24;
						type_code[2] = (map[y][x].graphic & 0x00FF0000) >> 16;
						piece_id[3] = (map[y][x].graphic & 0x0000FF00) >> 8;
						type_code[3] = (map[y][x].graphic & 0x000000FF);

						for(i=0;i<4;i++)
						{
							graphic[i] = FindTerrain(piece_id[i], type_code[i]);
							if (graphic[i] == 999)
								continue;

						  rect[i].left = tile[graphic[i]].x_origin;
						  rect[i].top = tile[graphic[i]].y_origin;
						  rect[i].right = tile[graphic[i]].x_extent;
						  rect[i].bottom = tile[graphic[i]].y_extent;

						  rectdst[i].left = x * 55 + view_x - tile[graphic[i]].x_offset;
						  rectdst[i].right = rectdst[i].left + (tile[graphic[i]].x_extent - tile[graphic[i]].x_origin);
						  rectdst[i].top = y * 55 + view_y - tile[graphic[i]].y_offset;
						  rectdst[i].bottom = rectdst[i].top + (tile[graphic[i]].y_extent - tile[graphic[i]].y_origin);
						}

            //Match up the draw ordering
            for(i=0;i<4;i++)
            {
              if (graphic[i] != 999)
              {
                switch(tile[graphic[i]].type_code)
                {
                case 0:
                  draworder[i] = 0;
                  break;
                case 8:
                  draworder[i] = 1;
                  break;
                case 7:
                  draworder[i] = 2;
                  break;
                case 4:
                  draworder[i] = 3;
                  break;
                case 1:
                  draworder[i] = 4;
                  break;
                case 2:
                  draworder[i] = 5;
                  break;
                case 6:
                  draworder[i] = 6;
                  break;
                case 5:
                  draworder[i] = 7;
                  break;
                default:
                  draworder[i] = 8;
                  break;
                }
              }
              else
                draworder[i] = 999;
            }
						

						//These graphic's need to be sorted so we draw the lowest draworder first
						for(i=0;i<4;i++)
						{
							int j;
							int draw = 0;
							//Find a tile that hasnt been drawn yet
							for(j=0;j<4;j++)
								if ( graphic[j] != 999)
                {
									draw = j;
                  break;
                }
							//Find the lowest draworder
							for(;j<4;j++)
								if (graphic[j] != 999)
                {
									if ( draworder[draw] > draworder[j])
										draw = j;
                }
							if (graphic[draw] != 999)
							{
                if (draworder[draw] < 5)
								  g_DDSurface->Blt( &rectdst[draw] ,g_DDImage, &rect[draw], DDBLT_KEYSRC | DDBLT_WAIT, NULL);
								graphic[draw] = 999; //Set the tile to not draw again
							}
						}
						
					}
		}
	}

  //Draw the x,y coords of the selected tile in the upper left of the screen
	g_DDSurface->GetDC( &dc );
	char stuff[20];
	itoa(selected_x, stuff, 10);
	TextOut(dc, 0, 20, stuff, strlen(stuff));
	itoa(selected_y, stuff, 10);
	TextOut(dc, 0, 40, stuff, strlen(stuff));
	g_DDSurface->ReleaseDC( dc );

	g_DDPrimarySurface->Flip(NULL, DDFLIP_WAIT);

	return true;
}